use log::debug;
use num::*;
use std::{
    fmt::{Debug, Display},
    ops::*,
};

pub mod direction;
pub mod map;
pub mod maximum;
mod test;
pub use direction::*;
pub use map::*;
pub use maximum::*;

pub type KD = KartesianDirection;

#[derive(Debug, Clone, Copy, Default, PartialEq, Eq, PartialOrd, Ord, Hash)]
pub struct Kartesian<T>
where
    T: Integer,
{
    pub x: T,
    pub y: T,
}

impl<T: Integer> Add for Kartesian<T> {
    fn add(self, rhs: Self) -> Self::Output {
        Kartesian { x: self.x + rhs.x, y: self.y + rhs.y }
    }

    type Output = Kartesian<T>;
}

impl<T: Integer + AddAssign> AddAssign for Kartesian<T> {
    fn add_assign(&mut self, rhs: Self) {
        self.x += rhs.x;
        self.y += rhs.y;
    }
}

impl<T: Integer> Sub for Kartesian<T> {
    fn sub(self, rhs: Self) -> Self::Output {
        Kartesian { x: self.x - rhs.x, y: self.y - rhs.y }
    }

    type Output = Kartesian<T>;
}

impl<T: Integer + SubAssign> SubAssign for Kartesian<T> {
    fn sub_assign(&mut self, rhs: Self) {
        self.x -= rhs.x;
        self.y -= rhs.y;
    }
}

impl<T: Integer> Kartesian<T> {
    pub const fn new(x: T, y: T) -> Self {
        Self { x: x, y: y }
    }
}

impl Kartesian<usize> {
    pub fn get_value<T: Copy>(self, map: &Vec<Vec<T>>) -> Option<T> {
        if map.len() > self.x {
            if map[self.x].len() > self.y {
                return Some(map[self.x][self.y]);
            }
        }

        None
    }
}

impl<T: Integer + Eq + Debug + CheckedAdd + CheckedSub + Default + Copy> Kartesian<T> {
    pub fn checked_add_max(self, rhs: Kartesian<T>, max: Kartesian<T>) -> Option<Kartesian<T>> {
        let mut new = Kartesian::default();
        new.x = match self.x.checked_add(&rhs.x) {
            None => return None,
            Some(x) => {
                if x < T::default() || x >= max.x {
                    return None;
                }
                x
            },
        };
        new.y = match self.y.checked_add(&rhs.y) {
            None => return None,
            Some(y) => {
                if y < T::default() || y >= max.y {
                    return None;
                }
                y
            },
        };
        Some(new)
    }
    pub fn checked_sub(self, rhs: Kartesian<T>) -> Option<Kartesian<T>> {
        let mut new = Kartesian::default();
        new.x = match self.x.checked_sub(&rhs.x) {
            None => return None,
            Some(x) => x,
        };
        new.y = match self.y.checked_sub(&rhs.y) {
            None => return None,
            Some(y) => y,
        };
        Some(new)
    }

    pub fn diff(self, rhs: Kartesian<T>) -> (Kartesian<T>, KartesianDirection) {
        let mut relative = Kartesian::<T>::default();
        let mut dir = KartesianDirection::None;
        debug!("{:?} <> {:?}", self.x, rhs.x);
        if self.x < rhs.x {
            dir = dir.down();
            relative.x = rhs.x - self.x;
        } else if self.x > rhs.x {
            dir = dir.up();
            relative.x = self.x - rhs.x;
        }
        debug!("{:?} <> {:?}", self.x, rhs.x);
        if self.y < rhs.y {
            dir = dir.right();
            relative.y = rhs.y - self.y;
        } else if self.y > rhs.y {
            dir = dir.left();
            relative.y = self.y - rhs.y;
        }
        (relative, dir)
    }

    #[inline]
    pub fn move_dir(self, vector: Kartesian<T>, dir: KartesianDirection) -> Option<Kartesian<T>> {
        let mut new = self;
        match dir {
            KartesianDirection::TopLeft | KartesianDirection::Top | KartesianDirection::TopRight => {
                new.x = match new.x.checked_sub(&vector.x) {
                    None => return None,
                    Some(d) => d,
                }
            },
            KartesianDirection::BottomLeft | KartesianDirection::Bottom | KartesianDirection::BottomRight => {
                new.x = match new.x.checked_add(&vector.x) {
                    None => return None,
                    Some(d) => d,
                }
            },
            _ => {},
        }
        match dir {
            KartesianDirection::TopLeft | KartesianDirection::Left | KartesianDirection::BottomLeft => {
                new.y = match new.y.checked_sub(&vector.y) {
                    None => return None,
                    Some(d) => d,
                }
            },
            KartesianDirection::TopRight | KartesianDirection::Right | KartesianDirection::BottomRight => {
                new.y = match new.y.checked_add(&vector.y) {
                    None => return None,
                    Some(d) => d,
                }
            },
            _ => {},
        }
        Some(new)
    }

    #[inline]
    pub fn move_dir_max(self, vector: Kartesian<T>, dir: KartesianDirection, max: Kartesian<T>) -> Option<Kartesian<T>> {
        match self.move_dir(vector, dir) {
            None => None,
            Some(new) => {
                if new.x < max.x && new.y < max.y {
                    Some(new)
                } else {
                    None
                }
            },
        }
    }
}

impl<T: Integer + Display + Copy> Display for Kartesian<T> {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        f.write_fmt(format_args!("{}:{}", self.x, self.y))
    }
}

impl<T: Integer + Copy> Div<T> for Kartesian<T> {
    type Output = Kartesian<T>;
    fn div(self, rhs: T) -> Self::Output {
        Kartesian {
            x: self.x.div_ceil(&rhs),
            y: self.y.div_ceil(&rhs),
        }
    }
}

impl<T: Integer + Rem> Rem for Kartesian<T> {
    type Output = Self;
    fn rem(self, rhs: Self) -> Self {
        Kartesian { x: self.x % rhs.x, y: self.y % rhs.y }
    }
}
impl<T: Integer + RemAssign> RemAssign for Kartesian<T> {
    fn rem_assign(&mut self, rhs: Self) {
        self.x %= rhs.x;
        self.y %= rhs.y;
    }
}

pub struct KartesianIterator {
    i: u8,
    diagonal: bool,
    none: bool,
    straight: bool,
}

impl Iterator for KartesianIterator {
    type Item = KartesianDirection;

    fn next(&mut self) -> Option<Self::Item> {
        let mut i = self.i;
        if self.straight {
            if i < 4 {
                self.i += 1;
            }
            match i {
                0 => return Some(KartesianDirection::Right),
                1 => return Some(KartesianDirection::Bottom),
                2 => return Some(KartesianDirection::Left),
                3 => return Some(KartesianDirection::Top),
                _ => i -= 4,
            }
        }
        if self.none {
            if i == 0 {
                self.i += 1;
                return Some(KartesianDirection::None);
            }
            i -= 1;
        }
        if self.diagonal {
            if i < 4 {
                self.i += 1;
            }
            match i {
                0 => return Some(KartesianDirection::TopLeft),
                1 => return Some(KartesianDirection::TopRight),
                2 => return Some(KartesianDirection::BottomLeft),
                3 => return Some(KartesianDirection::BottomRight),
                _ => {},
            }
        }
        None
    }
}
